-----------------------------------------------------------------------------
--
-- data
--  \threat
--    \<playerName>
--      - amount
--		\targetDetails
--        <targetHash> : amount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentThreat")
L:RegisterTranslations("enUS", function() return {
    ["Threat List"] = true,
} end)

local AssessmentThreat = Assessment:NewModule("Threat", "AceEvent-2.0")
local Threat = AceLibrary("Threat-1.0")

-----------------------------------------------------------------------------

Assessment.TagData.targetthreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end
				return Threat:GetThreat(data.key, targetName)
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.threat =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local threat = fightData.threat
				if not threat then return 0 end
				
				if not threat[data.key] then return 0 end
				return threat[data.key].amount					
			 end,
	data			= { "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.threattargetname =
{
	method = function(data) 
		if UnitExists("target") and not UnitIsFriend("player", "target") then
			return UnitName("target")
		end
	end,
}

Assessment.TagData.isusingforeignthreatsource =
{
	method = function(data) return data.key and Threat:IsUsingForeignThreatSource(data.key) end,
}

-----------------------------------------------------------------------------

Assessment.ListType.ThreatList =
{
    method = function(fightData)
                if not fightData.threat then return function() return nil end, nil, nil end
                return pairs(fightData.threat)
             end,
    data = "threat",
    description = L["Threat List"],
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets["Threat"] =
{
	titleTag		= "[threatTargetName:yellow][~threatTargetName:\"Threat\"]",
	listType		= "ThreatList", 
	barValueTag		= "select(threatTargetName, targetthreat, threat)", 
	filterType		= "NONZERO",
	fightIndex		= 1,
	barValueTextTag	= "[isUsingForeignThreatSource:\"*\"][barValue]",
}

Assessment.DefaultDisplaySets["TPS"] =
{
	titleTag		= "TPS: [combattime:format(\"%.1f\")]s",
	listType		= "ThreatList", 
	barValueTag		= "threat / maximum(combatTime, 1)", 
	filterType		= "NONZERO",
	fightIndex		= 1,
	barValueTextTag	= "[barValue:format(\"%.1f\")]",
}

-----------------------------------------------------------------------------

function AssessmentThreat:OnInitialize()
end

function AssessmentThreat:OnEnable()
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "OnPlayerTargetChanged")
	self:RegisterEvent("Threat_ThreatUpdated", "OnThreatUpdated")
end

function AssessmentThreat:OnDisable()
end

-----------------------------------------------------------------------------

function AssessmentThreat:OnPlayerTargetChanged()
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.target = GetTime()
	end
end

function AssessmentThreat:OnThreatUpdated(playerName, playerUnit, targetName, threatValue)
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.threat = GetTime()
		
		local threat = data.threat
		if not threat then
			threat = self:NewTable()
			data.threat = threat
		end
		
		local playerData = threat[playerName]
		if playerData then
			if targetName == "GLOBAL" and threatValue == 0 then
				threat[playerName] = self:ReleaseTable(playerData, true)
			else
				if threatValue == 0 then
					playerData.targetDetails[targetName] = nil
				else
					playerData.targetDetails[targetName] = threatValue
				end
				
				if threatValue >= playerData.amount then
					playerData.amount = threatValue
				else
					local maximumThreatValue = threatValue
				
					for k,v in pairs(playerData.targetDetails) do
						if v > maximumThreatValue then
							maximumThreatValue = v
						end
					end
					
					if maximumThreatValue == 0 then
						threat[playerName] = self:ReleaseTable(playerData, true)
					else
						playerData.amount = maximumThreatValue
					end
				end
			end
		elseif threatValue ~= 0 then
			playerData = self:NewTable()
			playerData.amount = threatValue
			playerData.targetDetails = self:NewTable()
			playerData.targetDetails[targetName] = threatValue
			threat[playerName] = playerData
		end
	end
end

-----------------------------------------------------------------------------
