-----------------------------------------------------------------------------
--
-- Per window configuration:
--  Title
--  Title appearance
--  Border
--  Background color
--  BarValue
--  SortCriteria
--  Text Tags
--  Bar appearance
--
-----------------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local SML = AceLibrary("SharedMedia-1.0")
local Roster = AceLibrary("Roster-2.1")
local Dewdrop = AceLibrary:HasInstance("Dewdrop-2.0") and AceLibrary("Dewdrop-2.0")

local SML_statusbar = SML:List("statusbar")
local SML_font = SML:List("font")

local AssessmentUserInterface = Assessment:NewModule("UserInterface", "AceEvent-2.0")

-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentUserInterface")
L:RegisterTranslations("enUS", function() return {
	["Windows"] = true,
	["Window options."] = true,
	["Display Sets"] = true,
	["Display set options."] = true,
	["Add New Window"] = true,
	["Add a new assessment window to the user interface."] = true,

	["Window %d"] = true,
	["Options for assessment window %d"] = true,
	["Display Set"] = true,
	["Choose what to show for this window"] = true,
	["Visible"] = true,
	["Make the assessment window visible"] = true,

	["Update Rate"] = true,
	["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."] = true,

	["Backdrop Color"] = true,
	["Change the backdrop color and alpha for the window."] = true,
	["Border Color"] = true,
	["Change the border color and alpha for the window."] = true,

	["Bar Options"] = true,
	["Bar options for this window"] = true,
	["Bar Count"] = true,
	["Number of bars to show"] = true,
	["Texture"] = true, 
	["Sets the bar texture"] = true,

	["Title Options"] = true,
	["Title options."] = true,
	["Enable"] = true,
	["Enable title text."] = true,
	["Alignment"] = true,
	["Select how to align the title."] = true,
	["Font"] = true,
	["Font for title."] = true,
	["Font Size"] = true,
	["Font size for title."] = true,
	["Font Shadow"] = true,
	["Show font shadow on title text"] = true,
	["Outline"] = true,
	["Sets the outline for title text"] = true,
	["Title Height"] = true,
	["Title height."] = true,

	["Remove"] = true,
	["Remove this window from display"] = true,

	["Add New Display Set"] = true,
	["Adds a new display set to the user interface."] = true,
	["Enter a new display set name"] = true,

	["Options for display set %s"] = true,

	["List Type"] = true,
	["Select a list type that controls what is displayed."] = true,
	["Bar Value Tag"] = true,
	["Select how to calculate bar values."] = true,
	["Sort Key Tag"] = true,
	["Select how to sort the bars."] = true,
	["Descending Sort"] = true,
	["Enable to use descending sort. Disable to use ascending."] = true,

	["Title Tag"] = true,
	["Select what to show for the window's title."] = true,

	["Filter Type"] = true,
	["Select how to filter the list."] = true,

	["FilterAll"] = "Show All",
	["FilterNonZero"] = "Show Non-Zero Values",

	["Fight Data"] = true,
	["Select which fight data set to display."] = true,
	["Total Data"] = true,
	["Current Fight"] = true,
	["Previous Fight"] = true,

	["Text"] = true,
	["Text options."] = true,
	["Add New Text"] = true,
	["Adds a new text to the display set."] = true,

	["Text %d"] = true,
	["Options for text %d"] = true,

	["Text Tag"] = true,
	["Select how to display text."] = true,

	["Alpha"] = true,
	["Set the text alpha."] = true,

	["Alignment"] = true,
	["Select how to align the text."] = true,
	["Left"] = true,
	["Center"] = true,
	["Right"] = true,

	["Font"] = true,
	["Font for text."] = true,

	["Font Size"] = true,
	["Font size for text."] = true,

	["Font Shadow"] = true,
	["Show font shadow on text"] = true,
	["Outline"] = true,
	["Sets the outline for text"] = true,

	["Remove"] = true,
	["Remove this text."] = true,

	["Remove"] = true,
	["Remove this display set."] = true,
	["This display set cannot be removed currently because one or more windows require its data."] = true,
	
	["Any Tag String"] = true,
	["None"] = true,
	["Normal"] = true,
	["Thick"] = true,
} end)

-----------------------------------------------------------------------------

AssessmentUserInterface.db = Assessment:AcquireDBNamespace("UserInterface")

local profileDefaults =
{
	windows =
	{
		count = 1,

		["**"] =
		{
			displaySetStack		= { "Damage Dealt" },
			displaySetStackLevel= 1,
			updateRate			= 20,
			title = 
			{
				enable			= true,
				alignment		= "CENTER",
				font			= "Arial Narrow",
				fontSize		= 9,
				outline			= "OUTLINE",
				shadow			= true,
				height			= 10,
				textColor		= { 1, 1, 1, 1 },
				backgroundColor	= { 0, 0, 0, 1 }
			},
			bar = 
			{ 
				count			= 10,
				texture			= "Smooth",
				classList		= { },
				colorByClass	= true,
				defaultColor	= { r = 0.75, g = 0.75, b = 0.75 },
			},
			frame = 
			{ 
				width			= 110,
				height			= 140, 
				point			= "CENTER",
				relativePoint	= "CENTER",
				x				= 0, 
				y				= 0, 
				borderColor		= { 1, 1, 1, 0.75 },
				backdropColor	= { 0, 0, 0, 0.75 },
				visible			= true,
			},
		},
	},
	displaySets =
	{
		-- Defaults are inserted here in OnInitialize
	
		["**"] =
		{
			titleTag		= "Assessment",
			
			listType		= "GroupMembers", 
			barValueTag		= "damageDealt", 
			sortKeyTag		= "barValue",
			filterType		= "NONZERO",
			descendingSort	= true,
			fightIndex		= 0,
			
			text =
			{
				count = 2,
				{
					tag			= "[rank]. [key]",
					offsets =
					{
						right	= -20,
					},
				},
				{
					tag			= "[barValue]",
					alignment	= "RIGHT",
				},
				["**"] = 
				{
					alignment	= "LEFT",
					font		= "Arial Narrow",
					fontSize	= 9,
					outline		= "",
					shadow		= true,
					alpha		= 1,
					offsets =
					{
						left		= 0,
						right		= 0,
						vertical	= 0,
					},
				}
			},
		},
	}
}

-----------------------------------------------------------------------------

Assessment.Options.args.uiSpacer =
{
	type = 'header',
	order = 50,
}

Assessment.Options.args.windows =
{
	type = "group",
	name = L["Windows"],
	desc = L["Window options."],
	args = 
	{ 
		add =
		{
			type = "execute",
			name = L["Add New Window"],
			desc = L["Add a new assessment window to the user interface."],
			func = function() AssessmentUserInterface:AddWindow() end,
			order = -1
		},
	},
	order = 60,
}

Assessment.Options.args.displaySets =
{
	type = "group",
	name = L["Display Sets"],
	desc = L["Display set options."],
	args = 
	{ 
		add =
		{
			type = "execute",
			name = L["Add New Display Set"],
			desc = L["Adds a new display set to the user interface."],
			func = function() AssessmentUserInterface:AddDisplaySet() end,
			order = -1
		},
	},
	order = 60,
}

-----------------------------------------------------------------------------

local function windowGet(passValue)
	local s = ("return function(t) return t[%d].%s end"):format(unpack(passValue))
	return loadstring(s)()(AssessmentUserInterface.db.profile.windows)
end

local function windowSet(passValue, value)
	local s = ("return function(t, v) t[%d].%s = v end"):format(unpack(passValue))
	loadstring(s)()(AssessmentUserInterface.db.profile.windows, value)
	AssessmentUserInterface:UpdateWindowOptions(passValue[1])
end

local function windowGetDisplaySet(passValue)
	local windowOptions = AssessmentUserInterface.db.profile.windows[passValue[1]]
	return windowOptions.displaySetStack[windowOptions.displaySetStackLevel]
end

local function windowSetDisplaySet(passValue, value)
	local windowOptions = AssessmentUserInterface.db.profile.windows[passValue[1]]
	
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack = { value }

	AssessmentUserInterface:UpdateWindowOptions(passValue[1])
end

local function displaySetGet(passValue)
	local s = ("return function(t) return t[\"%s\"].%s end"):format(unpack(passValue))
	local value = loadstring(s)()(AssessmentUserInterface.db.profile.displaySets)
	if passValue[2] == "fightIndex" then
		value = tostring(value)
	end
	return value
end

local function displaySetSet(passValue, value)
	if passValue[2] == "fightIndex" then
		value = tonumber(value)
	end

	local s = ("return function(t, v) t[\"%s\"].%s = v end"):format(unpack(passValue))
	loadstring(s)()(AssessmentUserInterface.db.profile.displaySets, value)
	AssessmentUserInterface:UpdateDisplaySets(passValue[1])
end

local function textGet(passValue)
	local s = ("return function(t) return t[\"%s\"].text[%d].%s end"):format(passValue[1][1], passValue[1][2], passValue[2])
	return loadstring(s)()(AssessmentUserInterface.db.profile.displaySets)
end

local function textSet(passValue, value)
	local s = ("return function(t, v) t[\"%s\"].text[%d].%s = v end"):format(passValue[1][1], passValue[1][2], passValue[2])
	loadstring(s)()(AssessmentUserInterface.db.profile.displaySets, value)
	AssessmentUserInterface:UpdateDisplaySets(passValue[1][1])
end

local function windowRemove(passValue)
	AssessmentUserInterface:RemoveWindow(passValue[1])
end

local function displaySetRemove(passValue)
	AssessmentUserInterface:RemoveDisplaySet(passValue[1])
end

-----------------------------------------------------------------------------

local displaySetValidate = { }
local listTypeValidate	 = { }
local filterTypeValidate = { }

-----------------------------------------------------------------------------

local windowOptionsTemplate =
{
	type = "group",
--	name = ["Window %d"] = true,
--	desc = ["Options for assessment window %d"] = true,
	args =
	{
		visible =
		{
			type = "toggle",
			name = L["Visible"],
			desc = L["Make the assessment window visible"],
			get = windowGet,
			set = windowSet,
			passValue = "frame.visible",
			order = 1,
		},
		backdropColor = 
		{
			type = "color",
			name = L["Backdrop Color"],
			desc = L["Change the backdrop color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.backdropColor",
			hasAlpha = true,
		},
		borderColor = 
		{
			type = "color",
			name = L["Border Color"],
			desc = L["Change the border color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.borderColor",
			hasAlpha = true,
		},
		displaySet = 
		{
			type = "text",
			name = L["Display Set"],
			desc = L["Choose what to show for this window"],
			get = windowGetDisplaySet,
			set = windowSetDisplaySet,
			passValue = "displaySet",
			validate = displaySetValidate,
			order = 2,
		},
		updateRate =
		{
			type = "range",
			name = L["Update Rate"],
			desc = L["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."],
			min  = 0,
			max  = 60,
			step = 1,
			passValue = "updateRate",
			get = windowGet,
			set = windowSet,
		},
		bars =
		{
			type = "group",
			name = L["Bar Options"],
			desc = L["Bar options for this window"],
			args =
			{
				count =
				{
					type = "range",
					name = L["Bar Count"],
					desc = L["Number of bars to show"],
					min	 = 1,
					max  = 50,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "bar.count",
					order = 1,
				},
				texture =
				{
					type = "text", 
					name = L["Texture"], 
					desc = L["Sets the bar texture"],
					get = windowGet,
					set = windowSet,
					validate = SML_statusbar,
					passValue = "bar.texture",
					order = 2,
				},
			},
		},
		title =
		{
			type = "group",
			name = L["Title Options"],
			desc = L["Title options."],
			args =
			{
				enable = 
				{
					type = "toggle",
					name = L["Enable"],
					desc = L["Enable title text."],
					get = windowGet,
					set = windowSet,
					passValue = "title.enable",
					order = 1,
				},
				alignment =
				{
					type = "text",
					name = L["Alignment"],
					desc = L["Select how to align the title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.alignment",
					validate = { LEFT = L["Left"], CENTER = L["Center"], RIGHT = L["Right"] },
				},
				font =
				{
					type = "text",
					name = L["Font"],
					desc = L["Font for title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.font",
					validate = SML_font,
				},
				height =
				{
					type = "range",
					name = L["Title Height"],
					desc = L["Title height."],
					min  = 0,
					max  = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.height",
				},
				fontSize =
				{
					type = "range",
					name = L["Font Size"],
					desc = L["Font size for title."],
					min = 4,
					max = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.fontSize",
				},
				shadow =
				{
					type = "toggle",
					name = L["Font Shadow"],
					desc = L["Show font shadow on title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.shadow",
				},
				outline =
				{
					type = "text",
					name = L["Outline"],
					desc = L["Sets the outline for title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.outline",
					validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
				},
			},
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this window from display"],
			func = windowRemove,
			passValue = false,
			order = -1,
		},
	}
}

local displaySetOptionsTemplate =
{
	type = "group",
--	name = ["%s"] = true,
--	desc = ["Options for display set %s"] = true,
	args =
	{
		listType =
		{
			type = "text",
			name = L["List Type"],
			desc = L["Select a list type that controls what is displayed."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "listType",
			validate = listTypeValidate,
			order = 1,
		},
		barValueTag =
		{
			type = "text",
			name = L["Bar Value Tag"],
			desc = L["Select how to calculate bar values."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "barValueTag",
			usage = L["Any Tag String"],
		},
		sortKeyTag =
		{
			type = "text",
			name = L["Sort Key Tag"],
			desc = L["Select how to sort the bars."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "sortKeyTag",
			usage = L["Any Tag String"],
		},
		titleTag =
		{
			type = "text",
			name = L["Title Tag"],
			desc = L["Select what to show for the window's title."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "titleTag",
			usage = L["Any Tag String"],
		},
		filterType =
		{
			type = "text",
			name = L["Filter Type"],
			desc = L["Select how to filter the list."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "filterType",
			validate = filterTypeValidate,
		},
		descendingSort =
		{
			type = "toggle",
			name = L["Descending Sort"],
			desc = L["Enable to use descending sort. Disable to use ascending."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "descendingSort",
		},
		fightData =
		{
			type = "text",
			name = L["Fight Data"],
			desc = L["Select which fight data set to display."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "fightIndex",
			validate = { ["0"] = L["Total Data"], ["1"] = L["Current Fight"], ["2"] = L["Previous Fight"] },
		},
		text =
		{
			type = "group",
			name = L["Text"],
			desc = L["Text options."],
			args =
			{
				add =
				{
					type = "execute",
					name = L["Add New Text"],
					desc = L["Adds a new text to the display set."],
					func = function(passValue) AssessmentUserInterface:AddText(passValue[1]) end,
					passValue = true,
					order = -1
				},
			},
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this display set."],
			func = displaySetRemove,
			passValue = false,
			order = -1,
		},
	}
}

local textOptionsTemplate =
{
	type = "group",
--	name = ["Text %d"] = true,
--	desc = ["Options for text %d"] = true,
	args =
	{
		tag =
		{
			type = "text",
			name = L["Text Tag"],
			desc = L["Select how to display text."],
			get = textGet,
			set = textSet,
			passValue = "tag",
			usage = L["Any Tag String"],
		},
		alignment =
		{
			type = "text",
			name = L["Alignment"],
			desc = L["Select how to align the text."],
			get = textGet,
			set = textSet,
			passValue = "alignment",
			validate = { LEFT = L["Left"], CENTER = L["Center"], RIGHT = L["Right"] },
		},
		alpha =
		{
			type = "range",
			name = L["Alpha"],
			desc = L["Set the text alpha."],
			min = 0,
			max = 1,
			step = 0.01,
			get = textGet,
			set = textSet,
			passValue = "alpha",
		},
		font =
		{
			type = "text",
			name = L["Font"],
			desc = L["Font for text."],
			get = textGet,
			set = textSet,
			passValue = "font",
			validate = SML_font,
		},
		fontSize =
		{
			type = "range",
			name = L["Font Size"],
			desc = L["Font size for text."],
			min = 4,
			max = 32,
			step = 1,
			get = textGet,
			set = textSet,
			passValue = "fontSize",
		},
		shadow =
		{
			type = "toggle",
			name = L["Font Shadow"],
			desc = L["Show font shadow on text"],
			get = textGet,
			set = textSet,
			passValue = "shadow",
		},
		outline =
		{
			type = "text",
			name = L["Outline"],
			desc = L["Sets the outline for text"],
			get = textGet,
			set = textSet,
			passValue = "outline",
			validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this text."],
			func = function(passValue) AssessmentUserInterface:RemoveText(unpack(passValue[1])) end,
			passValue = false,
			order = -1,
		},
	}
}

-----------------------------------------------------------------------------

Assessment.FilterType.ALL =
{
	method		= function() return true end,
	description	= L["FilterAll"],
}

Assessment.FilterType.NONZERO =
{
	method		= function(data) return data.barValue ~= 0 end,
	description	= L["FilterNonZero"],
}

-----------------------------------------------------------------------------

local UIBar = AceOO.Class()
function UIBar.prototype:init(parent, windowOptions, displaySet, index)
	UIBar.super.prototype.init(self)

	self.bar = parent:CreateTexture()
	self.texts = { }

	self:SetOptions(parent, windowOptions, displaySet, index)
end

function UIBar.prototype:SetOptions(parent, windowOptions, displaySet, index)
	self.windowOptions = windowOptions
	self.displaySet = displaySet

	local bar = self.bar
	bar:SetTexture(SML:Fetch("statusbar", windowOptions.bar.texture))
	bar:SetVertexColor(1, 1, 1, 1)
	bar:Hide()

	for i = 1, self.displaySet.text.count do
		if not self.texts[i] then
			self.texts[i] = parent:CreateFontString()
		end
	end

	for i = self.displaySet.text.count + 1, #self.texts do
		self.texts[i]:Hide()
	end
	
	self:PlaceBar(parent, index)
end

function UIBar.prototype:PlaceBar(parent, index)
	local bar = self.bar
	local verticalOffset = -5 - (self.windowOptions.title.enable and self.windowOptions.title.height or 0)
	local height = (parent:GetHeight()+verticalOffset-5) / self.windowOptions.bar.count
	
	bar:ClearAllPoints()
	bar:SetPoint("TOPLEFT", parent, "TOPLEFT", 5, -(index-1)*height+verticalOffset)
	bar:SetPoint("BOTTOMLEFT", parent, "TOPLEFT", 5, -index*height+verticalOffset)
	
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		local options = self.displaySet.text[i]
		
		text:ClearAllPoints()
		text:SetPoint("LEFT", parent, "LEFT", options.offsets.left+5, 0)
		text:SetPoint("RIGHT", parent, "RIGHT", options.offsets.right-5, 0)
		text:SetPoint("TOP", parent, "TOP", 0, options.offsets.vertical-(index-1)*height+verticalOffset)
		text:SetPoint("BOTTOM", parent, "TOP", 0, options.offsets.vertical-index*height+verticalOffset)

		text:SetAlpha(options.alpha)

		text:SetFont(SML:Fetch("font", options.font), options.fontSize, options.outline)
		if options.shadow then
			text:SetShadowOffset(1, -1)
		else
			text:SetShadowOffset(0, 0)
		end
		text:SetJustifyH(options.alignment)
	end
end

function UIBar.prototype:SetFraction(fraction)
	local width = fraction * (self.windowOptions.frame.width-10)
	if width == 0 then
		self.bar:Hide()
	else
		self.bar:SetWidth(width)
		self.bar:Show()
	end
end

function UIBar.prototype:Hide()
	self.bar:Hide()
	for i = 1, self.displaySet.text.count do
		self.texts[i]:Hide()
	end
end

function UIBar.prototype:GetBarColorForKey(name)
	local barOptions = self.windowOptions.bar

	if barOptions.colorByClass then
		local classList = barOptions.classList
		if not classList[name] then
			local unitid = Roster:GetUnitIDFromName(name)
			if not unitid then return end

			local _, class = UnitClass(unitid)
			classList[name] = class
			return RAID_CLASS_COLORS[class]
		end
		return RAID_CLASS_COLORS[ classList[name] ]
	end
end

function UIBar.prototype:Update(data, barNormalizeMultiplier, textMethods)
	self:SetFraction(data.barValue * barNormalizeMultiplier)
	
	local color = self:GetBarColorForKey(data.key)
	if not color then color = self.windowOptions.bar.defaultColor end
	
	self.bar:SetVertexColor(color.r, color.g, color.b)
	
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		text:SetText(textMethods[i](data))
		text:Show()
	end
end

-----------------------------------------------------------------------------

local UIWindow = AceOO.Class("AceEvent-2.0")

function UIWindow.prototype:init(windowOptions, displaySet)
	UIWindow.super.prototype.init(self)
	local frame = CreateFrame("Frame", nil, UIParent)
	frame.window = self

	self.frame			= frame
	self.bars			= { }
	
	self.titleText = frame:CreateFontString()
	self.titleTexture = frame:CreateTexture()

	local bottomRightDrag = CreateFrame("Frame", nil, frame)
	bottomRightDrag:Show()
	bottomRightDrag:SetFrameLevel( frame:GetFrameLevel() + 10)
	bottomRightDrag:SetWidth(16)
	bottomRightDrag:SetHeight(16)
	bottomRightDrag:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
	bottomRightDrag:EnableMouse(true)
	bottomRightDrag:SetScript("OnMouseDown", function() this:GetParent():StartSizing("BOTTOMRIGHT") end )
	bottomRightDrag:SetScript("OnMouseUp", function() this:GetParent():StopMovingOrSizing() end )

	self:SetWindowOptions(windowOptions)
end

function UIWindow.prototype:SetWindowOptions(windowOptions)
	local displaySet = AssessmentUserInterface:GetDisplaySet(windowOptions)

	self.dependentData		= { }
	self.dataUpdateTimes	= { }
	self.textMethods		= { }

	self.topBarOffset	= 0
	self.forceUpdate	= true
	self.displaySet		= displaySet
	self.windowOptions	= windowOptions

	self.listTypeMethod = Assessment.ListType[displaySet.listType].method
	self.filterMethod	= Assessment.FilterType[displaySet.filterType].method

	self:MergeRequiredData(Assessment.ListType[displaySet.listType].data)

	if displaySet.descendingSort then
		self.sortDirectionMethod = function(a, b) return a.sortValue > b.sortValue end
	else
		self.sortDirectionMethod = function(a, b) return a.sortValue < b.sortValue end
	end

	local compiledBarValue = Assessment:CompileTag(displaySet.barValueTag, "number")
	self.barValueMethod = compiledBarValue.method
	self:MergeRequiredData(compiledBarValue.data)
	
	for i=1,displaySet.text.count do
		local compiledTextValue = Assessment:CompileTag(displaySet.text[i].tag, "string")
		self.textMethods[i] = compiledTextValue.method
		self:MergeRequiredData(compiledTextValue.data)
	end

	local compiledSortValue = Assessment:CompileTag(displaySet.sortKeyTag, "raw")
	self.sortMethod = compiledSortValue.method

	-------

	local frame = self.frame
	local frameOptions = windowOptions.frame
	frame:SetWidth(frameOptions.width)
	frame:SetHeight(frameOptions.height)
	frame:SetPoint(frameOptions.point, UIParent, frameOptions.relativePoint, frameOptions.x, frameOptions.y)
	frame:SetMovable(true)
	frame:SetResizable(true)
	frame:EnableMouse(true)
	frame:SetMinResize(32,32)
	frame:EnableMouseWheel(true)

	frame:SetScript("OnMouseWheel", function(this, delta) 
										local window = this.window
										if delta < 0 then
											window.topBarOffset = window.topBarOffset + 1
										else
											window.topBarOffset = window.topBarOffset - 1
											if window.topBarOffset < 0 then
												window.topBarOffset = 0
											end
										end
										window.forceUpdate = true
										window:Update()
									end)
	
	frame:SetScript("OnMouseDown",		function(this) this:StartMoving() end)
	frame:SetScript("OnMouseUp",		function(this) 
											this:StopMovingOrSizing() 
											local point,_,relativePoint,x,y = this:GetPoint(1)
											local frameOptions = this.window.windowOptions.frame
											frameOptions.point = point
											frameOptions.relativePoint = relativePoint
											frameOptions.x = x
											frameOptions.y = y
										end)
	frame:SetScript("OnSizeChanged",	function(this, width, height) 
											local window = this.window
											local frameOptions = window.windowOptions.frame
											frameOptions.width = width 
											frameOptions.height = height
											window:LayoutBars()
											window.forceUpdate = true
											window:Update()
										end)

    frame:SetBackdrop({
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
        tile = true,
        tileSize = 16,
        edgeSize = 16,
		insets = {left = 5, right = 5, top = 5, bottom = 5}
    })  
	frame:SetBackdropBorderColor(unpack(windowOptions.frame.borderColor))
	frame:SetBackdropColor(unpack(windowOptions.frame.backdropColor))

	--------

	local titleText = self.titleText
	local titleOptions = windowOptions.title
	
	if titleOptions.enable then
		titleText:ClearAllPoints()
		titleText:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleText:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleText:SetTextColor(unpack(titleOptions.textColor))

		titleText:SetFont(SML:Fetch("font", titleOptions.font), titleOptions.fontSize, titleOptions.outline)
		if titleOptions.shadow then
			titleText:SetShadowOffset(1, -1)
		else
			titleText:SetShadowOffset(0, 0)
		end
		titleText:SetJustifyH(titleOptions.alignment)

		local compiledTitleTextValue = Assessment:CompileTag(displaySet.titleTag, "string")
		self.titleTextMethod = compiledTitleTextValue.method
		self:MergeRequiredData(compiledTitleTextValue.data)

		titleText:Show()
	else
		titleText:Hide()
	end

	--------
	
	local titleTexture = self.titleTexture
	titleTexture:ClearAllPoints()
	titleTexture:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
	titleTexture:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
	titleTexture:SetVertexColor(unpack(titleOptions.backgroundColor))

	--------

	for i = 1, windowOptions.bar.count do
		if self.bars[i] then
			self.bars[i]:SetOptions(frame, windowOptions, displaySet, i)
		else
			self.bars[i] = UIBar:new(frame, windowOptions, displaySet, i)
		end
	end
	for i = windowOptions.bar.count+1, #self.bars do
		self.bars[i]:Hide()
	end

	if frameOptions.visible then
		self:Show()
	else
		self:Hide()
	end
end

function UIWindow.prototype:Show()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
	end

	local windowOptions = self.windowOptions
	local updateRate = windowOptions.updateRate ~= 0 and 1/windowOptions.updateRate or 0

	self.updater = self:ScheduleRepeatingEvent(self.Update, updateRate, self)
	self.frame:Show()
	self:Update()
end

function UIWindow.prototype:Hide()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
		self.updater = nil
	end
	self.frame:Hide()
end

function UIWindow.prototype:LayoutBars(fightData)
	for i=1, self.windowOptions.bar.count do
		self.bars[i]:PlaceBar(self.frame, i)
	end
end

function UIWindow.prototype:MergeRequiredData(data)
	if not data then return end
	if type(data) == "string" then
		self.dependentData[data] = true
	elseif type(data) == "table" then
		for k in pairs(data) do
			self.dependentData[k] = true
		end
	end
end

function UIWindow.prototype:NeedsUpdate(fightData)
	local dataUpdateTimes = self.dataUpdateTimes

	if not fightData then
		if next(dataUpdateTimes) then
			for k in pairs(dataUpdateTimes) do
				dataUpdateTimes[k] = nil
			end
			return true
		end
		if self.forceUpdate then
			self.forceUpdate = nil
			return true
		end
		return false
	end

	local result = false
	for k in pairs(self.dependentData) do
		if dataUpdateTimes[k] ~= fightData.dataUpdateTimes[k] then
			result = true
			dataUpdateTimes[k] = fightData.dataUpdateTimes[k]
		end
	end
	if self.forceUpdate then
		self.forceUpdate = nil
		return true
	end
	return result
end

local sortList = { }
local titleData = { key = UnitName("player"), rank = 0 }
function UIWindow.prototype:DoUpdate(fightData)
	local maximumBarValue = 0
	local totalBarValue = 0

	for k in pairs(sortList) do
		sortList[k] = nil
	end

	if fightData then
		for k,v in self.listTypeMethod(fightData) do
			local data = Assessment:NewTable()
			data.key = k
			data.value = v
			data.fightData = fightData
			
			local barValue = self.barValueMethod(data)
			data.barValue = barValue
			
			if self.filterMethod(data) then
				if barValue > maximumBarValue then
					maximumBarValue = barValue
				end
				totalBarValue = totalBarValue + barValue

				data.barValue = barValue
				sortList[#sortList+1] = data
			end			
		end
		
		for _,data in pairs(sortList) do
			data.totalBarValue = totalBarValue
			data.maximumBarValue = maximumBarValue
			data.sortValue = self.sortMethod(data)
		end

		table.sort(sortList, self.sortDirectionMethod)
	end

	local barNormalizeMultiplier = maximumBarValue ~= 0 and 1 / maximumBarValue or 0

	if self.topBarOffset + self.windowOptions.bar.count > #sortList then
		self.topBarOffset = #sortList - self.windowOptions.bar.count
		if self.topBarOffset < 0 then
			self.topBarOffset = 0
		end
	end

	local hideStart = #sortList+1-self.topBarOffset
	if hideStart < 1 then hideStart = 1 end
	for i=hideStart, self.windowOptions.bar.count do
		self.bars[i]:Hide()
	end

	for i = 1+self.topBarOffset, #sortList do
		if i-self.topBarOffset > self.windowOptions.bar.count then break end
	
		sortList[i].rank = i
		self.bars[i-self.topBarOffset]:Update(sortList[i], barNormalizeMultiplier, self.textMethods)
		Assessment:ReleaseTable(sortList[i])
	end	

	if self.windowOptions.title.enable then
		titleData.fightData = fightData
		titleData.totalBarValue = totalBarValue
		titleData.maximumBarValue = maximumBarValue

		self.titleText:SetText( self.titleTextMethod(titleData) )
	end
end

function UIWindow.prototype:Update()
	local fightData = Assessment:GetFightData(self.displaySet.fightIndex)
	if not self:NeedsUpdate(fightData) then return end
	self:DoUpdate(fightData)
end

-----------------------------------------------------------------------------

function AssessmentUserInterface:OnInitialize()
	self.windows = { }

	for k,v in pairs(Assessment.DefaultDisplaySets) do
		local displaySet = { }
		profileDefaults.displaySets[k] = displaySet
		displaySet.titleTag		= v.titleTag
		displaySet.listType		= v.listType
		displaySet.barValueTag	= v.barValueTag
		displaySet.filterType	= v.filterType
		displaySet.fightIndex	= v.fightIndex
		if v.barValueTextTag then
			displaySet.text = { [2] = { tag = v.barValueTextTag }	}
		end
	end
	
	Assessment:RegisterDefaults("UserInterface", "profile", profileDefaults)
	
	for k,v in pairs(Assessment.ListType) do
		listTypeValidate[k] = v.description
	end

	for k,v in pairs(Assessment.FilterType) do
		filterTypeValidate[k] = v.description
	end
end

function AssessmentUserInterface:OnEnable()
	self:RegisterEvent('SharedMedia_SetGlobal', function(ty, val)
		if ty == "statusbar" then
			for k,v in ipairs(self.windows) do
				for k,v in ipairs(v.bars) do
					v.bar:SetTexture(SML:Fetch(ty, val))
				end
			end
		end
	end)
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentUserInterface:OnDisable()
end

function AssessmentUserInterface:CreateWindows()
	for i = 1, self.db.profile.windows.count do
		if not self.windows[i] then
			local windowOptions = self.db.profile.windows[i]
			self.windows[i] = UIWindow:new(windowOptions)
		end
	end
	
	for i = self.db.profile.windows.count+1, #self.windows do
		self.windows[i]:Hide()
	end
end

function AssessmentUserInterface:CreateOptions()
	for k in pairs(displaySetValidate) do
		displaySetValidate[k] = nil
	end

	for k in pairs(self.db.profile.displaySets) do
		displaySetValidate[k] = k
	end

	for i = 1, self.db.profile.windows.count do
		Assessment.Options.args.windows.args[i] = self:CreateWindowOptions(i)
	end
	
	for i = self.db.profile.windows.count+1, #Assessment.Options.args.windows.args do
		Assessment.Options.args.windows.args[i] = nil
	end

	for i = 1, #Assessment.Options.args.displaySets.args do
		Assessment.Options.args.displaySets.args[i] = nil
	end
	
	local i = 1
	for k in pairs(self.db.profile.displaySets) do
		Assessment.Options.args.displaySets.args[i] = self:CreateDisplaySetOptions(k)
		i = i + 1
	end
end

local function copyOptionsTemplate(source, i)
	local result = { }

	for k,v in pairs(source) do
		if type(v) == "table" then
			result[k] = copyOptionsTemplate(v, i)
		else
			if k == "passValue" then
				result[k] = { i, v }
			else
				result[k] = v
			end
		end
	end
	
	return result
end

function AssessmentUserInterface:CreateWindowOptions(i)
	local options = copyOptionsTemplate(windowOptionsTemplate, i)
	options.name = L["Window %d"]:format(i)
	options.desc = L["Options for assessment window %d"]:format(i)
	return options
end

function AssessmentUserInterface:CreateDisplaySetOptions(displaySetName)
	local options = copyOptionsTemplate(displaySetOptionsTemplate, displaySetName)
	options.name = displaySetName
	options.desc = L["Options for display set %s"]:format(displaySetName)

	for i = 1, self.db.profile.displaySets[displaySetName].text.count do
		options.args.text.args[i] = self:CreateTextOptions( { displaySetName, i } )
	end
	
	return options
end

function AssessmentUserInterface:CreateTextOptions(passValueData)
	local options = copyOptionsTemplate(textOptionsTemplate, passValueData)
	options.name = L["Text %d"]:format(passValueData[2])
	options.desc = L["Options for text %d"]:format(passValueData[2])
	return options
end

function AssessmentUserInterface:GetDisplaySet(windowOptions)
	return self.db.profile.displaySets[windowOptions.displaySetStack[windowOptions.displaySetStackLevel]]
end

function AssessmentUserInterface:AddWindow()
	local windowOptions = self.db.profile.windows
	local windows = self.windows
	
	windowOptions.count = windowOptions.count + 1
	
	if windows[windowOptions.count] then
		windows[windowOptions.count]:SetWindowOptions(windowOptions[windowOptions.count])
	else
		windows[windowOptions.count] = UIWindow:new(windowOptions[windowOptions.count])
	end

	self:CreateOptions()
end

function AssessmentUserInterface:RemoveWindow(index)
	local windowOptions = self.db.profile.windows
	local windows = self.windows

	local oldWindow = windows[index]
	local oldWindowOptions = windowOptions[index]

	for i = index, windowOptions.count do
		windowOptions[i] = windowOptions[i+1]
		windows[i] = windows[i+1]
	end
	windows[windowOptions.count] = oldWindow
	windowOptions[windowOptions.count] = oldWindowOptions
	
	oldWindow:Hide()
	
	windowOptions.count = windowOptions.count - 1
	
	self:CreateOptions()
end

function AssessmentUserInterface:UpdateWindowOptions(index)
	self.windows[index]:SetWindowOptions(self.db.profile.windows[index])
end

-----------------------------------------------------------------------------

StaticPopupDialogs["ASSESSMENT_ADD_DISPLAY_SET"] = 
{
	text = L["Enter a new display set name"],
	button1 = ACCEPT,
	button2 = CANCEL,
	hasEditBox = 1,
	OnAccept = function()
		local text = getglobal(this:GetParent():GetName().."EditBox"):GetText()
		AssessmentUserInterface:AddDisplaySet(text)
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus()
	end,
	EditBoxOnEnterPressed = function()
		local text = getglobal(this:GetParent():GetName().."EditBox"):GetText()
		AssessmentUserInterface:AddDisplaySet(text)
		this:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1,
}

function AssessmentUserInterface:AddDisplaySet(displaySetName)
	if displaySetName then
		self.db.profile.displaySets[displaySetName] = copyOptionsTemplate(profileDefaults.displaySets["**"])
		self:CreateOptions()
	else
		if Dewdrop then Dewdrop:Close() end
		StaticPopup_Show("ASSESSMENT_ADD_DISPLAY_SET")
	end
end

function AssessmentUserInterface:UpdateDisplaySets(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if windowOptions.displaySetStack[windowOptions.displaySetStackLevel] == displaySetName then
			self.windows[i]:SetWindowOptions(windowOptions)
		end
	end
end

function AssessmentUserInterface:RemoveDisplaySet(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		
		for j = 1, windowOptions.displaySetStackLevel do
			if windowOptions.displaySetStack[j] == displaySetName then
				Assessment:Print(L["This display set cannot be removed currently because one or more windows require its data."])
				return
			end
		end
	end
	
	self.db.profile.displaySets[displaySetName] = nil
	self:CreateOptions()
end

-----------------------------------------------------------------------------

function AssessmentUserInterface:AddText(displaySetName)
	local text = self.db.profile.displaySets[displaySetName].text
	text.count = text.count + 1

	self:UpdateDisplaySets(displaySetName)
	self:CreateOptions()
end

function AssessmentUserInterface:RemoveText(displaySetName, removeIndex)
	local text = self.db.profile.displaySets[displaySetName].text
	
	for i = removeIndex, text.count do
		text[i] = text[i+1]
	end
	
	text.count = text.count-1

	self:UpdateDisplaySets(displaySetName)

	self:CreateOptions()
end

-----------------------------------------------------------------------------
